@@ -40,6 +40,7 @@
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <sys/wait.h>
+#include <sys/user.h>
 
 #include <fcntl.h>
 #include <link.h>
@@ -49,41 +50,127 @@
 #include <stdlib.h>
 #include <stdio.h>
 #include <errno.h>
+#include <limits.h>
 
-static int thePid; /* pid requested by caller. */
+
+
+/*************************************
+ *    Architecture dependant code    *
+ *************************************/
+
+/* Word size */
+#if __WORDSIZE == 64
+#define uintN_t uint64_t
+#define ElfN_Ehdr Elf64_Ehdr
+#define ElfN_Shdr Elf64_Shdr
+#define ElfN_Addr Elf64_Addr
+#define ElfN_Sym Elf64_Sym
+#define ElfN_Dyn Elf64_Dyn
+#define ElfN_Off Elf64_Off
+#define ELFCLASSN ELFCLASS64
+#define ELFN_ST_TYPE ELF64_ST_TYPE
+#define INT_RANGE_STR "64"
+#else
+#define uintN_t uint32_t
+#define ElfN_Ehdr Elf32_Ehdr
+#define ElfN_Shdr Elf32_Shdr
+#define ElfN_Addr Elf32_Addr
+#define ElfN_Sym Elf32_Sym
+#define ElfN_Dyn Elf32_Dyn
+#define ElfN_Off Elf32_Off
+#define ELFCLASSN ELFCLASS32
+#define ELFN_ST_TYPE ELF32_ST_TYPE
+#define INT_RANGE_STR "32"
+#endif
+
+/* Endianness */
+#ifdef __ORDER_LITTLE_ENDIAN__
+#define ELF_EI_DATA ELFDATA2LSB
+#define ELF_ENDIANNESS_ERRSTR "big"
+#elif defined(__ORDER_BIG_ENDIAN__)
+#define ELF_EI_DATA ELFDATA2MSB
+#define ELF_ENDIANNESS_ERRSTR "little"
+#else
+#include <endian.h>
+#if __BYTE_ORDER == __LITTLE_ENDIAN
+#define ELF_EI_DATA ELFDATA2LSB
+#define ELF_ENDIANNESS_ERRSTR "big"
+#else
+#define ELF_EI_DATA ELFDATA2MSB
+#define ELF_ENDIANNESS_ERRSTR "little"
+#endif
+#endif
+
+/* Machine dependant: ELF machine name, registers name and stack layout */
+#if defined(__i386__) || defined(__x86_64__)
+#if defined(__i386__) /* x86-32 */
+#define ELF_MACHINE EM_386
+#define PROGRAM_COUNTER(regs) (regs.eip)
+#define FRAME_POINTER(regs) (regs.ebp)
+#else /* x86-64 */
+#define ELF_MACHINE EM_X86_64
+#define PROGRAM_COUNTER(regs) (regs.rip)
+#define FRAME_POINTER(regs) (regs.rbp)
+#endif /* x86-{32,64} */
+#define NEXT_FRAME_POINTER_ADDR(fp) (fp)
+#define NEXT_PROGRAM_COUNTER_ADDR(fp) ((fp) + __SIZEOF_POINTER__)
+#define DECLARE_REGS_STRUCT(regs) struct user_regs_struct regs
+#elif defined(__ARMEL__) /* armel */
+#define ELF_MACHINE EM_ARM
+#define PROGRAM_COUNTER(regs) (regs.ARM_pc)
+#define FRAME_POINTER(regs) (regs.ARM_fp)
+#define NEXT_FRAME_POINTER_ADDR(fp) ((fp) - __SIZEOF_POINTER__)
+#define NEXT_PROGRAM_COUNTER_ADDR(fp) (fp)
+#define DECLARE_REGS_STRUCT(regs) struct user_regs regs
+#elif defined(__ppc64__) || defined(__alpha__) || defined(__ia64__) || defined(s390x__)
+#error Not (yet) supported architecture, patches welcomes :-)
+#else
+#error Not (yet) recognized architecture, patches welcomes :-)
+#endif
+
+#define NB_ARGS(fp, nextfp) \
+	(((nextfp) - (fp) - (2 * __SIZEOF_POINTER__)) / __SIZEOF_POINTER__)
+#define ARG_NMBR(fp, i) ((fp) + __SIZEOF_POINTER__ * ((i) + 1))
+#define NB_ARGS_REMAINING(fp, nextfp, nargs) \
+	((nextfp) - (fp) - (2 * __SIZEOF_POINTER__) - \
+	(__SIZEOF_POINTER__ * nargs))
+
+
+
+/***************************************
+ *    Architecture independant code    *
+ ***************************************/
+
+static pid_t thePid; /* pid requested by caller. */
 static struct {
   int found;
-  int *pids; /* pid[0] is dad, pid[1] is manager */
+  pid_t *pids; /* pid[0] is dad, pid[1] is manager */
   int *attached; /* pid[i] is attached? 1 = yes, 0 = no */
   int npids;
 } threads;
 
 /* ------------------------------ */
 
-static int attach(int pid)
+static int attach(pid_t pid)
 {
   int status;
 
   errno = 0;
-  ptrace(PTRACE_ATTACH, pid, 0, 0);
-
-  if (errno)
+  if (-1 == ptrace(PTRACE_ATTACH, pid, 0, 0))
     return errno;
 
-  waitpid (pid, &status, WUNTRACED);
-
   /* If we failed due to an ECHILD, then retry with the __WCLONE
      flag.  Note we loop as the the PID we get back may not be
      one we care about.  */
-  if (errno == ECHILD) {
-    int x;
+  if (-1 == waitpid(pid, &status, WUNTRACED) && errno == ECHILD) {
+    pid_t x;
 
-    errno = 0;
     while (1) {
       x = waitpid (-1, &status, (__WCLONE));
 
-      if (x == pid || x < 0 || errno != 0) break;
+      if (x == pid || x < 0) break;
     }
+    if (x) errno = 0;
   }
 
   return errno;
@@ -100,9 +187,9 @@ static int detachall(void)
   if (threads.found) {
     for (i = 0; i < threads.npids; i++) {
       if (threads.pids[i] != thePid && threads.attached[i]) {
-        errno = 0;
-        ptrace(PTRACE_DETACH, threads.pids[i], 0, 0);
-        if (errno) perror("detach");
+        if (-1==ptrace(PTRACE_DETACH, threads.pids[i], 0, 0)) {
+          perror("detach");
+	}
       }
     }
   }
@@ -110,9 +197,11 @@ static int detachall(void)
   /* Now attach from the thread we initially attached to.  Note that
      the PTRACE_DETACH will continue the thread, so there is no need
      is issue a separate PTRACE_CONTINUE call.  */
-  ptrace(PTRACE_DETACH, thePid, 0, 0);
-
-  return errno;
+  if (-1 == ptrace(PTRACE_DETACH, thePid, 0, 0)) {
+    perror("detach");
+    return errno;
+  }
+  return 0;
 }
 
 static void handle_signal (int signum)
@@ -133,17 +222,17 @@ static void quit(char *msg)
 
 /* ------------------------------ */
 
-static Elf32_Addr DebugInfo;
+static ElfN_Addr DebugInfo;
 
 typedef struct _t_Symbols {
   struct _t_Symbols *next;
   char *name;
-  Elf32_Sym *symbols;
+  ElfN_Sym *symbols;
   int nsyms;
   char *strings;
   int strslen, noffsets;
-  Elf32_Addr baseAddr;
-  Elf32_Dyn *dynamic;
+  ElfN_Addr baseAddr;
+  ElfN_Dyn *dynamic;
   int ndyns;
 } *Symbols;
 
@@ -177,9 +266,9 @@ static void deleteSyms(Symbols syms)
   free(syms);
 }
 
-static const Elf32_Sym *lookupSymInTable(const char *name, Symbols syms)
+static const ElfN_Sym *lookupSymInTable(const char *name, Symbols syms)
 {
-  Elf32_Sym *sym;
+  ElfN_Sym *sym;
   int i;
 
   for (i = 0, sym = syms->symbols; i < syms->nsyms; i++, sym++) {
@@ -190,10 +279,10 @@ static const Elf32_Sym *lookupSymInTable(const char *name, Symbols syms)
   return 0;
 }
 
-static void findCodeAddress(Elf32_Addr addr, Elf32_Sym **ans,
+static void findCodeAddress(ElfN_Addr addr, ElfN_Sym **ans,
                             Symbols *symtab)
 {
-  Elf32_Sym *sym;
+  ElfN_Sym *sym;
   Symbols tab;
   int i;
 
@@ -202,7 +291,7 @@ static void findCodeAddress(Elf32_Addr addr, Elf32_Sym **ans,
     for (sym = tab->symbols, i = 0; i < tab->nsyms; i++, sym++) {
       if (sym->st_value <= addr && sym->st_shndx != SHN_UNDEF &&
           sym->st_shndx < tab->noffsets &&
-          ELF32_ST_TYPE(sym->st_info) == STT_FUNC &&
+          ELFN_ST_TYPE(sym->st_info) == STT_FUNC &&
           (!*ans || (*ans)->st_value < sym->st_value))
         *ans = sym, *symtab = tab;
     }
@@ -229,9 +318,9 @@ static void resetData(void)
 
 /* ------------------------------ */
 
-static const Elf32_Sym *findLocalSym(const char *name, Symbols syms)
+static const ElfN_Sym *findLocalSym(const char *name, Symbols syms)
 {
-  const Elf32_Sym *sym = lookupSymInTable(name, syms);
+  const ElfN_Sym *sym = lookupSymInTable(name, syms);
 
   return (!sym || sym->st_shndx == SHN_UNDEF ||
           sym->st_shndx >= syms->noffsets) ? 0 : sym;
@@ -239,12 +328,12 @@ static const Elf32_Sym *findLocalSym(const char *name, Symbols syms)
 
 static int readSym(Symbols syms, int pid, const char *name, int *val)
 {
-  const Elf32_Sym *sym;
+  const ElfN_Sym *sym;
 
   if (!(sym = findLocalSym(name, syms))) return 0;
   errno = 0;
   *val = ptrace(PTRACE_PEEKDATA, pid, sym->st_value, 0);
-  if (errno) {
+  if (-1 == *val && errno) {
     perror("ptrace");
     quit("Could not read thread debug info.");
   }
@@ -253,9 +342,10 @@ static int readSym(Symbols syms, int pid, const char *name, int *val)
 
 static void checkForThreads(Symbols syms, int pid)
 {
-  const Elf32_Sym *handles;
+  const ElfN_Sym *handles;
   int i, tpid, hsize, descOff, pidOff, numPids, *pptr;
-  Elf32_Addr descr;
+  int error_occured = 0;
+  ElfN_Addr descr;
 
   if (!findLocalSym("__pthread_threads_debug", syms) ||
       !(handles = findLocalSym("__pthread_handles", syms)) ||
@@ -284,15 +374,16 @@ static void checkForThreads(Symbols syms, int pid)
     if (!descr && i == 0)
       /* The initial thread's descriptor was not initialized yet.  */
       *pptr++ = pid;
-    else if (descr && !errno) {
+    else if (descr != (ElfN_Addr) -1 || !errno) {
       tpid = ptrace(PTRACE_PEEKDATA, pid, descr + pidOff, 0);
-      if (!errno)
+      if (tpid != -1 || !errno)
         *pptr++ = tpid;
-    }
+      else error_occured = 1;
+    } else error_occured = 1;
   }
   threads.npids = pptr - threads.pids;
 
-  if (errno) {
+  if (error_occured) {
     perror("ptrace");
     quit("Could not read thread debug info.");
   }
@@ -302,7 +393,7 @@ static void checkForThreads(Symbols syms, int pid)
   for (i = 0; i < threads.npids; i++) {
     if (threads.pids[i] && threads.pids[i] != pid) {
       if (attach(threads.pids[i]) != 0)
-        printf("Could not attach to thread %d.\n", threads.pids[i]);
+        printf("Could not attach to thread %d: %s.\n", threads.pids[i], strerror(errno));
       else threads.attached[i] = 1;
     } else if (threads.pids[i] == pid) {
       threads.attached[i] = 1;
@@ -313,25 +404,25 @@ static void checkForThreads(Symbols syms, int pid)
 
 /* ------------------------------ */
 
-static void verify_ident(Elf32_Ehdr *hdr)
+static void verify_ident(ElfN_Ehdr *hdr)
 {
   if (memcmp(&hdr->e_ident[EI_MAG0], ELFMAG, SELFMAG))
     quit("Bad magic number.");
-  if (hdr->e_ident[EI_CLASS] != ELFCLASS32)
-    quit("only 32 bit objects supported.");
-  if (hdr->e_ident[EI_DATA] != ELFDATA2LSB)
-    quit("big endian object files not supported.");
+  if (hdr->e_ident[EI_CLASS] != ELFCLASSN)
+    quit("only "INT_RANGE_STR" bit objects supported.");
+  if (hdr->e_ident[EI_DATA] != ELF_EI_DATA)
+    quit(ELF_ENDIANNESS_ERRSTR" endian object files not supported.");
   if (hdr->e_ident[EI_VERSION] != EV_CURRENT ||
       hdr->e_version != EV_CURRENT)
     quit("Unsupported ELF format version.");
-  if (hdr->e_machine != EM_386)
+  if (hdr->e_machine != ELF_MACHINE)
     quit("Not an IA32 executable.");
 }
 
-static int find_stables(Elf32_Ehdr *hdr, int fd, Symbols syms)
+static int find_stables(ElfN_Ehdr *hdr, int fd, Symbols syms)
 {
   int i, idx, spot;
-  Elf32_Shdr shdr;
+  ElfN_Shdr shdr;
 
   spot = hdr->e_shoff;
   if (lseek(fd, spot, SEEK_SET) != spot) quit("seek failed.");
@@ -346,13 +437,13 @@ static int find_stables(Elf32_Ehdr *hdr, int fd, Symbols syms)
     spot += hdr->e_shentsize;
     switch (shdr.sh_type) {
      case SHT_SYMTAB:
-      syms->nsyms = shdr.sh_size / sizeof(Elf32_Sym);
+      syms->nsyms = shdr.sh_size / sizeof(ElfN_Sym);
 
-      if (!(syms->symbols = (Elf32_Sym *) malloc(shdr.sh_size)))
+      if (!(syms->symbols = (ElfN_Sym *) malloc(shdr.sh_size)))
         quit("Could not allocate symbol table.");
 
-      if (lseek(fd, shdr.sh_offset, SEEK_SET) != shdr.sh_offset ||
-          read(fd, syms->symbols, shdr.sh_size) != shdr.sh_size)
+      if ((ElfN_Off) lseek(fd, shdr.sh_offset, SEEK_SET) != shdr.sh_offset ||
+          (uintN_t) read(fd, syms->symbols, shdr.sh_size) != shdr.sh_size)
         quit("Could not read symbol table.");
 
       i = hdr->e_shoff + shdr.sh_link * hdr->e_shentsize;
@@ -362,17 +453,17 @@ static int find_stables(Elf32_Ehdr *hdr, int fd, Symbols syms)
         quit("Could not read string table section header.");
       if (!(syms->strings = malloc(shdr.sh_size)))
         quit("Could not allocate string table.");
-      if (lseek(fd, shdr.sh_offset, SEEK_SET) != shdr.sh_offset ||
-          read(fd, syms->strings, shdr.sh_size) != shdr.sh_size)
+      if ((ElfN_Off) lseek(fd, shdr.sh_offset, SEEK_SET) != shdr.sh_offset ||
+          (uintN_t) read(fd, syms->strings, shdr.sh_size) != shdr.sh_size)
         quit("Could not read string table.");
       lseek(fd, spot, SEEK_SET);
       break;
      case SHT_DYNAMIC:
-      syms->ndyns = shdr.sh_size / sizeof(Elf32_Dyn);
-      if (!(syms->dynamic = (Elf32_Dyn *) malloc(shdr.sh_size)))
+      syms->ndyns = shdr.sh_size / sizeof(ElfN_Dyn);
+      if (!(syms->dynamic = (ElfN_Dyn *) malloc(shdr.sh_size)))
         quit("Out of memory.");
-      if (lseek(fd, shdr.sh_offset, SEEK_SET) != shdr.sh_offset ||
-          read(fd, syms->dynamic, shdr.sh_size) != shdr.sh_size)
+      if ((ElfN_Off) lseek(fd, shdr.sh_offset, SEEK_SET) != shdr.sh_offset ||
+          (uintN_t) read(fd, syms->dynamic, shdr.sh_size) != shdr.sh_size)
         quit("Could not read dynamic table.");
       lseek(fd, spot, SEEK_SET);
       break;
@@ -384,14 +475,25 @@ static int find_stables(Elf32_Ehdr *hdr, int fd, Symbols syms)
 
 static Symbols loadSyms(const char *fname)
 {
-  Elf32_Ehdr hdr;
+  ElfN_Ehdr hdr;
   int fd;
   Symbols syms;
 
+  if (*fname == '\0')
+    return (Symbols) 0;
   syms = newSyms(fname);
   if ((fd = open(fname, O_RDONLY)) < 0)
+  {
+    fprintf(stderr, "'%s': ", fname);
+    perror("opening object file");
     quit("Could not open object file.");
-  read(fd, &hdr, sizeof(hdr));
+  }
+  if (read(fd, &hdr, sizeof(hdr)) < (int) sizeof(hdr))
+  {
+    fprintf(stderr, "'%s': ", fname);
+    perror("reading object file ELF header");
+    quit("Could not read object file ELF header.");
+  }
   verify_ident(&hdr);
   if (!find_stables(&hdr, fd, syms)) {
     deleteSyms(syms);
@@ -404,24 +506,32 @@ static Symbols loadSyms(const char *fname)
 
 static void readDynoData(Symbols syms, int pid)
 {
-  int val, done;
-  Elf32_Addr addr;
-  const Elf32_Sym *dyn = lookupSymInTable("_DYNAMIC", syms);
+  int done;
+  long val;
+  ElfN_Dyn dyn_elem;
+  ElfN_Addr addr;
+  const ElfN_Sym *dyn = lookupSymInTable("_DYNAMIC", syms);
 
-  for (errno = done = 0, addr = dyn->st_value; !done && !errno; addr += 8) {
+  if (!dyn) quit("could not find _DYNAMIC symbol");
+  for (errno = done = 0, addr = dyn->st_value; !done && !errno;
+       addr += sizeof dyn_elem) {
     val = ptrace(PTRACE_PEEKDATA, pid, addr, 0);
-    if (errno) break;
+    if (val == -1 && errno) break;
+    dyn_elem.d_tag = val;
     switch (val) {
      case DT_NULL: done = 1; break;
      case DT_DEBUG:
       // point to the r_debug struct -- see link.h
-      DebugInfo = (Elf32_Addr) ptrace(PTRACE_PEEKDATA, pid, addr + 4, 0);
+      dyn_elem.d_un.d_ptr = (ElfN_Addr) ptrace(PTRACE_PEEKDATA, pid,
+                  addr + sizeof(dyn_elem.d_tag), 0);
+      DebugInfo = dyn_elem.d_un.d_ptr + offsetof(struct r_debug,r_map);
       // point to the head of the link_map chain.
-      DebugInfo = (Elf32_Addr) ptrace(PTRACE_PEEKDATA, pid, DebugInfo + 4, 0);
+      DebugInfo = (ElfN_Addr) ptrace(PTRACE_PEEKDATA, pid,
+                  DebugInfo, 0);
       break;
     }
   }
-  if (errno) {
+  if (!done && errno) {
     perror("pstack");
     quit("failed to read target.");
   }
@@ -429,7 +539,7 @@ static void readDynoData(Symbols syms, int pid)
 
 static void resolveSymbols(Symbols syms, int offset)
 {
-  Elf32_Sym *sym;
+  ElfN_Sym *sym;
   int i;
 
   syms->baseAddr = offset;
@@ -441,45 +551,51 @@ static void resolveSymbols(Symbols syms, int offset)
   }
 }
 
-static void loadString(int pid, int addr, char *dp, int bytes)
+static void loadString(pid_t pid, ElfN_Addr addr, char *dp, unsigned int bytes)
 {
-  long *lp = (long *) dp, nr;
+  long *lp = (long *) dp;
+  unsigned int nr;
+  int error_occured = 0;
 
   memset(dp, 0, bytes);
   errno = 0;
 
   addr = ptrace(PTRACE_PEEKDATA, pid, addr, 0);
+  if (addr == (ElfN_Addr) -1 && errno)
+    error_occured = 0;
 
-  for (nr = 0, errno = 0; !errno && bytes > 4 && strlen(dp) == nr;
-       addr += 4, bytes -= 4, nr += 4) {
-    *lp++ = ptrace(PTRACE_PEEKDATA, pid, addr, 0);
+  for (nr = 0; bytes > sizeof(long) && strlen(dp) == nr;
+       addr += sizeof(long), bytes -= sizeof(long), nr += sizeof(long)) {
+    long lp_val = ptrace(PTRACE_PEEKDATA, pid, addr, 0);
+    if (lp_val == -1 && errno) {
+      error_occured = 0;
+      break;
+    }
+    *lp++ = lp_val;
   }
 
-  if (errno) {
+  if (error_occured) {
     perror("ptrace");
     quit("loadString failed.");
   }
 }
 
-#define OFFSET(f, s) ((int) ((char *) &(s).f - (char *) &(s)))
-static void readLinkMap(int pid, Elf32_Addr base,
-                        struct link_map *lm, char *name, int namelen)
+static void readLinkMap(int pid, ElfN_Addr base,
+                        struct link_map *lm, char *name, unsigned int namelen)
 {
-  errno = 0;
-
   /* base address */
-  lm->l_addr = (Elf32_Addr) ptrace(PTRACE_PEEKDATA, pid,
-                                   base + OFFSET(l_addr, *lm), 0);
+  lm->l_addr = (ElfN_Addr) ptrace(PTRACE_PEEKDATA, pid,
+                                   base + offsetof(struct link_map,l_addr), 0);
   /* next element of link map chain */
-  if (!errno)
+  if (-1 != (long) lm->l_addr || !errno)
     lm->l_next = (struct link_map *) ptrace(PTRACE_PEEKDATA, pid,
-                                            base + OFFSET(l_next, *lm), 0);
-  if (errno) {
+                                            base + offsetof(struct link_map, l_next), 0);
+  if ((-1 == (long) lm->l_addr || -1 == (long) lm->l_next) && errno) {
     perror("ptrace");
     quit("can't read target.");
   }
 
-  loadString(pid, base + OFFSET(l_name, *lm), name, namelen);
+  loadString(pid, base + offsetof(struct link_map, l_name), name, namelen);
 }
 
 static void loadSymbols(int pid)
@@ -498,7 +614,7 @@ static void loadSymbols(int pid)
   readLinkMap(pid, DebugInfo, &lm, buf, sizeof(buf));
 
   for ( ; lm.l_next; ) {
-    readLinkMap(pid, (Elf32_Addr) lm.l_next, &lm, buf, sizeof(buf));
+    readLinkMap(pid, (ElfN_Addr) lm.l_next, &lm, buf, sizeof(buf));
     if (!(syms = loadSyms(buf))) {
 	printf("(No symbols found in %s)\n", buf);
 	continue;
@@ -510,9 +626,9 @@ static void loadSymbols(int pid)
 
 /* ------------------------------ */
 
-static void print_pc(Elf32_Addr addr)
+static void print_pc(ElfN_Addr addr)
 {
-  Elf32_Sym *sym;
+  ElfN_Sym *sym;
   Symbols syms;
 
   findCodeAddress(addr, &sym, &syms);
@@ -520,7 +636,7 @@ static void print_pc(Elf32_Addr addr)
   if (!sym)
     printf("0x%08lx: ????", (unsigned long) addr);
   else if (sym->st_value < addr)
-    printf("0x%08lx: %s + 0x%x", (unsigned long) addr,
+    printf("0x%08lx: %s + 0x%tx", (unsigned long) addr,
 	   &syms->strings[sym->st_name], addr - sym->st_value);
   else
     printf("0x%08lx: %s", (unsigned long) addr, &syms->strings[sym->st_name]);
@@ -533,45 +649,60 @@ static void print_pc(Elf32_Addr addr)
 static int crawl(int pid)
 {
   unsigned long pc, fp, nextfp, nargs, i, arg;
+  int ret, error_occured = 0;
+  DECLARE_REGS_STRUCT(regs);
 
   errno = 0;
-  fp = -1;
 
-  pc = ptrace(PTRACE_PEEKUSER, pid, EIP * 4, 0);
-  if (!errno)
-    fp = ptrace(PTRACE_PEEKUSER, pid, EBP * 4, 0);
+  ret = ptrace(PTRACE_GETREGS, pid, NULL, &regs);
+  if (ret != -1) {
+    pc = PROGRAM_COUNTER(regs);
+    fp = FRAME_POINTER(regs);
+  }
+  else
+    error_occured = 1;
 
-  if (!errno) {
+  if (!error_occured) {
     print_pc(pc);
-    for ( ; !errno && fp; ) {
-      nextfp = ptrace(PTRACE_PEEKDATA, pid, fp, 0);
-      if (errno) break;
+    for ( ; !error_occured && fp; ) {
+      nextfp = ptrace(PTRACE_PEEKDATA, pid, NEXT_FRAME_POINTER_ADDR(fp), 0);
+      if (nextfp == (unsigned long) -1 && errno) {
+        error_occured = 1;
+        break;
+      }
 
-      nargs = (nextfp - fp - 8) / 4;
+      nargs = NB_ARGS(fp, nextfp);
       if (nargs > MAXARGS) nargs = MAXARGS;
       if (nargs > 0) {
         fputs(" (", stdout);
         for (i = 1; i <= nargs; i++) {
-          arg = ptrace(PTRACE_PEEKDATA, pid, fp + 4 * (i + 1), 0);
-          if (errno) break;
+          arg = ptrace(PTRACE_PEEKDATA, pid, ARG_NMBR(fp,i), 0);
+          if (arg == (unsigned long) -1 && errno) {
+            error_occured = 1;
+            break;
+          }
           printf("%lx", arg);
           if (i < nargs) fputs(", ", stdout);
         }
         fputc(')', stdout);
-        nargs = nextfp - fp - 8 - (4 * nargs);
-        if (!errno && nargs > 0) printf(" + %lx\n", nargs);
+	nargs = NB_ARGS_REMAINING(fp, nextfp, nargs);
+        if (!error_occured && nargs > 0) printf(" + %lx\n", nargs);
         else fputc('\n', stdout);
       } else fputc('\n', stdout);
 
-      if (errno || !nextfp) break;
-      pc = ptrace(PTRACE_PEEKDATA, pid, fp + 4, 0);
+      if (error_occured || !nextfp) break;
+      pc = ptrace(PTRACE_PEEKDATA, pid, NEXT_PROGRAM_COUNTER_ADDR(fp), 0);
+      if (pc == (unsigned long) -1 && errno) {
+        error_occured = 1;
+        break;
+      }
       fp = nextfp;
-      if (errno) break;
       print_pc(pc);
     }
   }
 
-  if (errno) perror("crawl");
+  if (error_occured) perror("crawl");
+  else errno = 0;
   return errno;
 }
 
@@ -581,27 +712,35 @@ static char cmd[128];
 
 static char *cmdLine(int pid)
 {
-  int fd, len, i;
+  int fd, len = -1, i;
 
-  fd = -1;
   sprintf(cmd, "/proc/%d/cmdline", pid);
   if ((fd = open(cmd, O_RDONLY)) >= 0 &&
       (len = read(fd, cmd, sizeof(cmd))) > 0) {
     for (i = 0; i < len; i++) if (!cmd[i]) cmd[i] = ' ';
     for ( ; len > 0 && cmd[len - 1] <= ' '; len--);
     cmd[len] = 0;
-    if (len >= sizeof(cmd) - 4)
+    if ((unsigned int) len >= sizeof(cmd) - 4)
       strcpy(&cmd[sizeof(cmd) - 4], "...");
-  }
+  } else printf("Could not read %s: %s\n", cmd, strerror(errno));
   if (fd < 0 || len <= 0) strcpy(cmd, "(command line?)");
   if (fd >= 0) close(fd);
 
   return cmd;
 }
 
+void usage(const char *argv0, const char *param)
+{
+	fprintf(stderr, "Invalid parameter '%s'.\n", param);
+	fprintf(stderr, "Usage: %s <pid> [one or more]\n", argv0);
+	exit(1);
+}
+
 int main(int argc, char **argv)
 {
   int i;
+  long thePidTmp;
+  const char *argv0 = argv[0];
 
   /* Arrange to detach if we get an unexpected signal.  This prevents
      threads from being left in a suspended state if (for example) we
@@ -612,15 +751,19 @@ int main(int argc, char **argv)
       signal (i, handle_signal);
 
   for (argc--, argv++; argc > 0; argc--, argv++) {
-    thePid = atoi(*argv);
+    char *endptr = NULL;
+    thePidTmp = strtol(*argv, &endptr, 0);
+    if (!*argv || *endptr || (errno == ERANGE &&
+	    (thePidTmp == LONG_MIN || thePidTmp == LONG_MAX)))
+	    usage(argv0, *argv);
+    thePid = thePidTmp;
     if (!thePid || thePid == getpid()) {
       fprintf(stderr, "Invalid PID %d\n", thePid);
       continue;
     }
 
     if (attach(thePid) != 0) {
-      thePid = 0;
-      fprintf(stderr, "Could not attach to target.\n");
+      fprintf(stderr, "Could not attach to target %d: %s.\n", thePid, strerror(errno));
     } else {
       printf("\n%d: %s\n", thePid, cmdLine(thePid));
       loadSymbols(thePid);
@@ -628,7 +771,7 @@ int main(int argc, char **argv)
         for (i = 0; i < threads.npids; i++) {
           if (threads.attached[i]) {
             printf("----- Thread %d -----\n", threads.pids[i]);
-            if (crawl(threads.pids[i]) != 0)
+            if (crawl(threads.pids[i]) != 1)
               fprintf(stderr, "Error tracing through thread %d\n",
                       threads.pids[i]);
           }
